﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections.Concurrent;
using System.Threading.Tasks;
using System.IO;
using Microsoft.Concurrency.TestTools.UnitTesting;
using System.Text.RegularExpressions;

namespace AlpacaProject.ArchitecturalPatterns
{
    public class ArchitecturalPatternsTests
    {

        [UnitTestMethod]
        public void TestProducerConsumer()
        {
            int buffersize = 1024;
            int num_producers = 4;
            int num_consumers = 4;

            // create bounded buffer
            var buffer = new BlockingCollection<IDValueItem>(buffersize);

            // start producers
            Task[] producers = new Task[num_producers];
            for (int i = 0; i < num_producers; i++)
                producers[i] = Task.Factory.StartNew(() => new NewItemProducer<IDValueItem>(buffer, 3).Run());

            // Once all the producers are finished producing, let the buffer know its done.
            Task.Factory.ContinueWhenAll(producers, _ => buffer.CompleteAdding());

            // start consumers
            Task[] consumers = new Task[num_consumers];
            for (int i = 0; i < num_consumers; i++)
                consumers[i] = Task.Factory.StartNew(() => new GenericItemConsumer<IDValueItem>(buffer).Run());

            Task.WaitAll(producers);
            Task.WaitAll(consumers);
        }

        [UnitTestMethod]
        public void TestPipeline()
        {
            int buffersize = 1024;
            int stage1_NumWorkers = 4;
            int stage2_NumWorkers = 4;
            int stage3_NumWorkers = 4;

            // create bounded buffers
            var buffer_1_2 = new BlockingCollection<IDValueItem>(buffersize);
            var buffer_2_3 = new BlockingCollection<IDValueItem>(buffersize);

            // Start stage 1 workers
            var stage1Workers = new Task[stage1_NumWorkers];
            for (int i = 0; i < stage1_NumWorkers; i++)
                stage1Workers[i] = Task.Factory.StartNew(() =>
                    new NewItemProducer<IDValueItem>(buffer_1_2, 10) {
                        Name = "Stage1"
                    }.Run());
            // Once all the stage 1 workers are finished, let the buffer know we're done producing stuff.
            Task.Factory.ContinueWhenAll(stage1Workers, _ => buffer_1_2.CompleteAdding());

            // Start stage 2 workers
            var stage2Workers = new Task[stage2_NumWorkers];
            for (int i = 0; i < stage2_NumWorkers; i++)
                stage2Workers[i] = Task.Factory.StartNew(() =>
                    new LogItemStage<IDValueItem>(buffer_1_2, buffer_2_3) {
                        Name = "Stage2"
                    }.Run());
            // Once all the stage 2 workers are finished, let the buffer know we're done producing stuff.
            Task.Factory.ContinueWhenAll(stage2Workers, _ => buffer_2_3.CompleteAdding());

            // Start stage 3 workers
            var stage3Workers = new Task[stage3_NumWorkers];
            for (int i = 0; i < stage3_NumWorkers; i++)
                stage3Workers[i] = Task.Factory.StartNew(() =>
                    new GenericItemConsumer<IDValueItem>(buffer_2_3) {
                        Name = "Stage3"
                    }.Run());

            // Wait for them to all finish
            Task.WaitAll(stage1Workers);
            Task.WaitAll(stage2Workers);
            Task.WaitAll(stage3Workers);
        }

        [UnitTestMethod]
        public void TestWorkList_TraverseFileSystem()
        {
            // This example will find all the *.csproj files on the current hard drive.
            // Note, this is a manual traversal of the tree and is used only for
            // demonstration purposes. A better way to search would just be to use
            // DirectoryInfo.GetFiles(searchPattern, SearchOption.AllDirectories)

            BlockingCollection<FileInfo> foundFiles = new BlockingCollection<FileInfo>();

            var worklist = new WorkList<DirectoryInfo>(GetCurrentRootFolders());
            worklist.ProcessWork = dir => {
                //Console.WriteLine(dir.FullName);

                try
                {
                    // Look for all proj files
                    var projFiles = dir.EnumerateFiles("*.csproj");

                    // If none exist, than traverse deeper...
                    if (!projFiles.Any())
                    {
                        worklist.AddWork(dir.EnumerateDirectories());
                        return;
                    }

                    // Now that we have the proj file, handle it
                    foundFiles.AddFromEnumerable(projFiles, false);
                }
                catch (UnauthorizedAccessException)
                { }
            };
            worklist.Run();

            int cnt = 0;
            foreach (var projFile in foundFiles)
            {
                cnt++;
                Console.WriteLine(projFile.FullName);
            }
            Console.WriteLine("Total found: {0}", cnt);
        }

        private IEnumerable<DirectoryInfo> GetCurrentRootFolders()
        {
            string windowsFolderPath = Environment.GetFolderPath(Environment.SpecialFolder.Windows, Environment.SpecialFolderOption.None);

            var rootDir = new DirectoryInfo(Environment.CurrentDirectory).Root;
            return from dir in rootDir.GetDirectories()
                   where !dir.FullName.StartsWith(windowsFolderPath)
                   select dir
                   ;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="textToFindRegexPattern"></param>
        /// <param name="fileSearchPatterns">
        /// WARNING: 
        /// </param>
        [UnitTestMethod]
        [TestArgs(@"\bTItem\b", "*.cs")]
        public void CompositionTest_FindFilesContainingText(string textToFindRegexPattern, params string[] fileSearchPatterns)
        {
            // Note, this is a manual traversal of the tree and is used only for
            // demonstration purposes. A better way to search would just be to use
            // DirectoryInfo.GetFiles(searchPattern, SearchOption.AllDirectories)

            // Make sure the patterns are valid
            if (fileSearchPatterns == null || fileSearchPatterns.Length == 0)
                throw new ArgumentNullException("fileSearchPattern", "Must be specified and more than one.");

            // Create the regex obj we'll be using
            Regex textToFindRegex = new Regex(textToFindRegexPattern, RegexOptions.Compiled);

            // Create the func that utilizes the search patterns for use in the ProcessWork action.
            Func<DirectoryInfo, IEnumerable<string>> enumerateFileInDir = 
                dir => fileSearchPatterns
                    .SelectMany(searchPattern => dir.EnumerateFiles(searchPattern))
                    .Select(fi => fi.FullName)
                    ;

            // 1st Stage: Find the files matching the fileSearchPatterns
            var filesToRead = new BlockingCollection<string>(1000);
            var stage1Worker = Task.Factory.StartNew(() => {
                try
                {
                    var findFilesWorklist = new WorkList<DirectoryInfo>(GetCurrentRootFolders());
                    findFilesWorklist.ProcessWork = dir => {
                        //Console.WriteLine(dir.FullName);

                        try
                        {
                            // Look for all files that match any of the patterns
                            filesToRead.AddFromEnumerable(enumerateFileInDir(dir), false);

                            // continue traversing the directory structure
                            findFilesWorklist.AddWork(dir.EnumerateDirectories());
                        }
                        catch (UnauthorizedAccessException)
                        { }
                    };
                    findFilesWorklist.Run();
                }
                finally
                {
                    // Let our buffer know we're finally done.
                    filesToRead.CompleteAdding();
                }
            });

            // 2nd Stage: Open files and perform a Regex search on the text
            var resultFiles = new BlockingCollection<string>();
            var stage2Workers = PatternFactory.StartPipelineStageWorkers<string, string>(4
                , filesToRead, resultFiles
                , filePath => {
                    //Console.WriteLine("Stage 2: {0}", filePath);
                    using (TextReader reader = new StreamReader(filePath))
                    {
                        if (textToFindRegex.IsMatch(reader.ReadToEnd()))
                            return filePath;
                        else
                            return null;    // Filtering on this will have to be done later.
                    }
                });
            Task.Factory.ContinueWhenAll(stage2Workers, _ => resultFiles.CompleteAdding());

            // 3rd Stage: Consume the resultFiles
            int cnt = 0;
            var stage3Worker = PatternFactory.StartConsumer(resultFiles, filePath => {
                // Filter files that didn't match
                if (filePath == null) return;

                System.Threading.Interlocked.Increment(ref cnt);
                Console.WriteLine("Stage 3: {0}", filePath);
            });

            stage3Worker.Wait();

            Console.WriteLine("Total found: {0}", cnt);
        }

    }
}
