﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections.Concurrent;
using Microsoft.Concurrency.TestTools.UnitTesting;
using System.Threading.Tasks;

namespace AlpacaProject.ArchitecturalPatterns
{
    public class Provisioning
    {

        [UnitTestMethod]
        public void ManualProvisioningExample()
        {
            int buffersize = 1024;
            int N = 1000;

            BlockingCollection<IDValueItem> buffer = new BlockingCollection<IDValueItem>(buffersize);

            // Create the producer
            Task producer = PatternFactory.StartNewItemProducer<IDValueItem>(buffer, N, true);

            int count = 0;
            FixedConsumers(buffer, _ => System.Threading.Interlocked.Increment(ref count));
            Assert.AreEqual(N, count);
        }

        private void FixedConsumers<TItem>(BlockingCollection<TItem> buffer, Action<TItem> consume)
        {
            int num_workers = Environment.ProcessorCount;

            Task[] workers = new Task[num_workers];
            for (int i = 0; i < num_workers; i++)
            {
                workers[i] = Task.Factory.StartNew(() => {
                    foreach (var item in buffer.GetConsumingEnumerable())
                        consume(item);
                });
            }

            Task.WaitAll(workers);
        }

        [UnitTestMethod]
        public void FlexibleProvisioningExample()
        {
            int buffersize = 1024;
            int N = 1000;

            BlockingCollection<IDValueItem> buffer = new BlockingCollection<IDValueItem>(buffersize);

            // Create the producer
            Task producer = PatternFactory.StartNewItemProducer<IDValueItem>(buffer, N, true);

            int count = 0;
            FlexibleConsumers(buffer, _ => System.Threading.Interlocked.Increment(ref count));
            Assert.AreEqual(N, count);
        }

        private void FlexibleConsumers<TItem>(BlockingCollection<TItem> buffer, Action<TItem> consume)
        {
            // Here we use the BlockingCollection.GetConsumingPartitioner
            // extension method found in the ParallelExtensionsExtras assembly
            // See: http://blogs.msdn.com/b/pfxteam/archive/2010/04/06/9990420.aspx

            Parallel.ForEach(
                buffer.GetConsumingPartitioner()
                , consume
                );
        }

        private TaskMeter fixedMeter = new TaskMeter("Fixed");
        private TaskMeter flexibleMeter = new TaskMeter("Flexible");
        private TaskMeter createBufferMeter = new TaskMeter("Create Buffer");

        private Task StartNewItemProducer<TItem>(BlockingCollection<TItem> buffer, int cnt)
            where TItem : class, new()
        {
            return Task.Factory.StartNew(() => {
                try
                {
                    createBufferMeter.Start();
                    new NewItemProducer<TItem>(buffer, 1000).Run();
                }
                finally
                {
                    createBufferMeter.End();
                    buffer.CompleteAdding();
                }
            });

        }

        [PerformanceTestMethod(WarmupRepetitions = 2, Repetitions = 10)]
        [TestArgs(true)]
        [TestArgs(false)]
        public void PerformanceTest(bool waitForProducersToFinish)
        {
            int N = 1000;
            int buffersize = Math.Min(N, 1024);

            // Create the producer
            BlockingCollection<IDValueItem> buffer1 = new BlockingCollection<IDValueItem>(buffersize);
            Task producer1 = StartNewItemProducer(buffer1, N);
            BlockingCollection<IDValueItem> buffer2 = new BlockingCollection<IDValueItem>(buffersize);
            Task producer2 = StartNewItemProducer(buffer2, N);
            if (waitForProducersToFinish)
            {
                Assert.IsTrue(N <= buffersize, "When setting waitForProducersToFinish to true, we need N to be less than or equal to the buffersize, otherwise Task.Wait will block for forever because the consumers won't have a chance to handle items in the collection.");
                Task.WaitAll(producer1, producer2);
            }

            // Time FixedConsumers
            {
                int count = 0;
                fixedMeter.Start();
                FixedConsumers(buffer1, _ => System.Threading.Interlocked.Increment(ref count));
                fixedMeter.End();
                Assert.AreEqual(N, count);
            }

            // Time FlexibleConsumers
            {
                int count = 0;
                flexibleMeter.Start();
                FlexibleConsumers(buffer2, _ => System.Threading.Interlocked.Increment(ref count));
                flexibleMeter.End();
                Assert.AreEqual(N, count);
            }
        }

    }
}
