﻿//--------------------------------------------------------------------------
// 
//  Copyright (c) Microsoft Corporation.  All rights reserved. 
// 
//  File: RoomPoint.cs
//
//--------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Microsoft.Concurrency.TestTools.Extensions;

namespace AntisocialRobots
{
    public class RobotSimulation : RobotSimulationBase
    {

        public enum FrameUpdateMode
        {
            Sequential = 0,
            ParallelNoLocks,
            Parallel_WithBigLock,
            /// <summary>
            /// This version  will lock too much and cause a cycle in the lock acquisition graph
            /// </summary>
            Parallel_WithFineGrainedLocks1,
            Parallel_WithFineGrainedLocks2,
            Parallel_WithLocks_Final,
            StrippedParallel,
        }

        // for fixing the bug(s)
        object _cellsLock;
        object[,] _cellLocks;

        public RobotSimulation(int size)
            : base(size)
        {
            _cellsLock = new object();

            _cellLocks = new Object[ROOM_SIZE, ROOM_SIZE];
            for (int i = 0; i < ROOM_SIZE; i++)
                for (int j = 0; j < ROOM_SIZE; j++)
                    _cellLocks[i, j] = new Object();
        }

        public FrameUpdateMode UpdateMode { get; set; }
        public bool EnableSleepDuringSimulateOneStep { get; set; }

        protected override void OnPerformFrameUpdate(int frameIdx)
        {
            switch (UpdateMode)
            {
                case FrameUpdateMode.Sequential:
                    FrameUpdate_Sequential();
                    break;
                case FrameUpdateMode.ParallelNoLocks:
                    FrameUpdate_Parallel(SimulateOneStep_NoLocks);
                    break;
                case FrameUpdateMode.Parallel_WithBigLock:
                    FrameUpdate_Parallel(SimulateOneStep_Parallel_WithBigLock);
                    break;
                case FrameUpdateMode.Parallel_WithFineGrainedLocks1:
                    FrameUpdate_Parallel(SimulateOneStep_WithFineGrainedLocks1);
                    break;
                case FrameUpdateMode.Parallel_WithFineGrainedLocks2:
                    FrameUpdate_Parallel(SimulateOneStep_WithFineGrainedLocks2);
                    break;
                case FrameUpdateMode.Parallel_WithLocks_Final:
                    FrameUpdate_Parallel(SimulateOneStep_WithLocks_Final);
                    break;
                case FrameUpdateMode.StrippedParallel:
                    FrameUpdate_Stripped(frameIdx);
                    break;
                default:
                    throw new NotImplementedException("FrameUpdateMode not handled: " + UpdateMode);
            }
        }

        private void FrameUpdate_Sequential()
        {
            Action<Robot> SimulateOneStep = SimulateOneStep_NoLocks;
            foreach (Robot robot in _movableRobots)
                SimulateOneStep(robot);
        }

        private void FrameUpdate_Parallel(Action<Robot> SimulateOneStep)
        {
            FauxParallel.ForEach(_movableRobots, SimulateOneStep);
        }

        private void FrameUpdate_Stripped(int frameIndex)
        {
            Action<Robot> SimulateOneStep = SimulateOneStep_NoLocks;

            for (int stripe = 0; stripe < 9; stripe++)
                FauxParallel.ForEach(_movableRobots, (Robot r) => {
                    if (r.lastmoved < frameIndex
                        && (r.Location.X % 3) == (stripe % 3)
                        && (r.Location.Y % 3) == (stripe / 3))
                    {
                        SimulateOneStep(r);
                        r.lastmoved = frameIndex;
                    }
                });
        }

        /// <summary>Performs one step of the simulation for one robot.</summary>
        /// <param name="r">The robot to perform the simulation step for.</param>
        private void SimulateOneStep_NoLocks(Robot r)
        {
            if (!r.IsMovable)
                return;

            // When doing the load test, things happen so fast with only two movable robots
            // that it's really hard get the bug to present itself. So by adding this sleep
            // here, we enable a preemption that allows for the interaction between the two
            // robots to happen.
            if (EnableSleepDuringSimulateOneStep)
                System.Threading.Thread.Sleep(1);

            RoomPoint origLoc = r.Location;
            RoomPoint newLoc = DetermineNewLocation(r);

            //** if the robot is moving and the location is available
            if (!newLoc.Equals(origLoc))
            {
                if (_roomCells[newLoc.X, newLoc.Y] == null)
                {
                    // a move of a robot from its cell to an adjoining cell is not
                    // guaranteed to be atomic
                    _roomCells[origLoc.X, origLoc.Y] = null;
                    _roomCells[newLoc.X, newLoc.Y] = r;
                    // the following is not guaranteed to be an atomic update
                    // because RoomPoint is a struct !!!
                    r.Location = newLoc;
                }
            }
        }

        /// <summary>
        /// Computes the new desired location for the robot.
        /// This version doesn't do any thread synchronization
        /// operations, and is therefore NOT thread-safe.
        /// </summary>
        /// <param name="r"></param>
        /// <returns></returns>
        private RoomPoint DetermineNewLocation(Robot r)
        {
            // Set ptR to the location of the robot in room coordinates
            // Note that RoomPoint is a struct!
            RoomPoint ptR = r.Location;
            double vectorX = 0, vectorY = 0;

            foreach (Robot s in _robots)
            {
                if (r == s) continue;
                RoomPoint ptS = s.Location;

                double inverseSquareDistance = 1.0 / RoomPoint.Square(ptR.DistanceTo(ptS));
                double angle = ptR.AngleTo(ptS);
                vectorX -= inverseSquareDistance * Math.Cos(angle);
                vectorY -= inverseSquareDistance * Math.Sin(angle);
            }

            return ComputeNewLocation(ptR, vectorX, vectorY, ROOM_SIZE);
        }

        /// <summary>Performs one step of the simulation for one robot.</summary>
        /// <param name="r">The robot to perform the simulation step for.</param>
        private void SimulateOneStep_Parallel_WithBigLock(Robot r)
        {
            if (!r.IsMovable)
                return;

            RoomPoint origLoc = r.Location;
            RoomPoint newLoc = DetermineNewLocation(r);

            // if the robot is moving and the location is available
            if (!newLoc.Equals(r.Location))
            {
                // Here, we use a big lock that all threads use. This causes
                // us to loose some parallelism because of the communication between the
                // threads to obtain this lock.
                lock (_cellsLock)
                {
                    if (_roomCells[newLoc.X, newLoc.Y] == null)
                    {
                        _roomCells[newLoc.X, newLoc.Y] = r;
                        _roomCells[r.Location.X, r.Location.Y] = null;
                        lock (r)
                            r.Location = newLoc;
                    }
                }
            }
        }

        /// <summary>Performs one step of the simulation for one robot.</summary>
        /// <param name="r">The robot to perform the simulation step for.</param>
        private void SimulateOneStep_WithFineGrainedLocks1(Robot r)
        {
            if (!r.IsMovable)
                return;

            if (EnableSleepDuringSimulateOneStep)
                System.Threading.Thread.Sleep(1);

            RoomPoint origLoc = r.Location;

            // Protect access to other robots so we can focus just on the deadlock below
            RoomPoint newLoc = DetermineNewLocation_ThreadSafe(r);

            // if the robot is moving and the location is available
            if (!newLoc.Equals(origLoc))
            {
                // Deadlock Scenario:
                //   When two robots adjacent to each other would like to
                //   move to the other robot's cell.
                // Why? Because the order of acquiring the locks may be out of order
                // from robot to robot.
                lock (_cellLocks[newLoc.X, newLoc.Y])
                {
                    lock (_cellLocks[origLoc.X, origLoc.Y])
                    {
                        if (_roomCells[newLoc.X, newLoc.Y] == null)
                        {
                            _roomCells[newLoc.X, newLoc.Y] = r;
                            _roomCells[origLoc.X, origLoc.Y] = null;
                            lock (r)
                                r.Location = newLoc;
                        }
                    }
                }

            }
        }

        /// <summary>Performs one step of the simulation for one robot.</summary>
        /// <param name="r">The robot to perform the simulation step for.</param>
        private void SimulateOneStep_WithFineGrainedLocks2(Robot r)
        {
            if (!r.IsMovable)
                return;

            if (EnableSleepDuringSimulateOneStep)
                System.Threading.Thread.Sleep(1);

            RoomPoint origLoc = r.Location;

            // Protect access to other robots so we can focus just on the deadlock below
            RoomPoint newLoc = DetermineNewLocation_ThreadSafe(r);

            // if the robot is moving and the location is available
            if (!newLoc.Equals(origLoc))
            {
                // Deadlock Fix:
                //   Acquire locks in a consistent order.
                //   Here, I'm just choosing to use the location of the cell in a 1D view
                //   of the array and using it's index.
                object newCellLock = _cellLocks[newLoc.X, newLoc.Y];
                object origCellLock = _cellLocks[origLoc.X, origLoc.Y];
                object cellLockLow, cellLockHi;
                if ((newLoc.Y * ROOM_SIZE + newLoc.X) < (origLoc.Y * ROOM_SIZE + origLoc.X))
                {
                    cellLockLow = newCellLock;
                    cellLockHi = origCellLock;
                }
                else
                {
                    cellLockLow = origCellLock;
                    cellLockHi = newCellLock;
                }

                lock (cellLockLow)
                {
                    lock (cellLockHi)
                    {
                        if (_roomCells[newLoc.X, newLoc.Y] == null)
                        {
                            _roomCells[newLoc.X, newLoc.Y] = r;
                            _roomCells[origLoc.X, origLoc.Y] = null;
                            lock (r)
                                r.Location = newLoc;
                        }
                    }
                }
            }
        }

        /// <summary>Performs one step of the simulation for one robot.</summary>
        /// <param name="r">The robot to perform the simulation step for.</param>
        private void SimulateOneStep_WithLocks_Final(Robot r)
        {
            if (!r.IsMovable)
                return;

            if (EnableSleepDuringSimulateOneStep)
                System.Threading.Thread.Sleep(1);

            RoomPoint origLoc = r.Location;

            // Protect access to other robots so we can focus just on the deadlock below
            RoomPoint newLoc = DetermineNewLocation_ThreadSafe(r);

            // if the robot is moving and the location is available
            if (!newLoc.Equals(origLoc))
            {
                // Alternative Locking Scheme:
                //   Note, we need to keep the lock for setting r.Location.
                //   For this scheme, we prevent obtaining the lock on the original
                //   location until we know we'll need it due to proceeding with
                //   the move.
                //   But it's still a question of whether there's a scenario that
                //   may cause a deadlock with this scheme since we aren't ordering
                //   the locks anymore.
                //   The answer lies in the result of the if condition in this
                //   algorithm.
                //   
                object newCellLock = _cellLocks[newLoc.X, newLoc.Y];
                object origCellLock = _cellLocks[origLoc.X, origLoc.Y];

                lock (newCellLock)
                {
                    if (_roomCells[newLoc.X, newLoc.Y] == null)
                    {
                        _roomCells[newLoc.X, newLoc.Y] = r;

                        // Still need to lock this for other threads that may be racing
                        // between lock(newCellLock) for this cell and the if condition.
                        // Can you come up with a scenario where without this lock
                        // we'd have a data race and/or deadlock?
                        // Data Race: Where are the reads/writes?
                        // Deadlock: Find an interleaving where two threads can reach
                        //    this point with the same two cells.
                        lock (origCellLock)
                            _roomCells[origLoc.X, origLoc.Y] = null;

                        // Robot instances are protected separately than cells
                        lock (r)
                            r.Location = newLoc;
                    }
                }
            }
        }

        private RoomPoint DetermineNewLocation_ThreadSafe(Robot r)
        {
            // Set ptR to the location of the robot in room coordinates
            // Note that RoomPoint is a struct!
            RoomPoint ptR = r.Location;
            double vectorX = 0, vectorY = 0;

            foreach (Robot s in _robots)
            {
                if (r == s) continue;
                RoomPoint ptS;
                lock (s)
                    ptS = s.Location;

                double inverseSquareDistance = 1.0 / RoomPoint.Square(ptR.DistanceTo(ptS));
                double angle = ptR.AngleTo(ptS);
                vectorX -= inverseSquareDistance * Math.Cos(angle);
                vectorY -= inverseSquareDistance * Math.Sin(angle);
            }

            ptR = ComputeNewLocation(ptR, vectorX, vectorY, ROOM_SIZE);
            return ptR;
        }

    }
}
